#!/usr/bin/perl -w
use Tk; # for GUI

# takes a table file and computes the average of some high bands (e.g. 3rd to 7th)
# and of some low bands (from 60th on) -> ratio between them
# from the ratio we find the coefficient to find true bands

# names for clones: lib.*plateNumber.*well.*.fsa
# table name: lib.*plateNumber.*.txt
# both according to distances in parameter.cfg

my $currColor; # current color
my $prevColor; # previous color
my $currWell;
my $prevWell;
my $currPlate;
my $prevPlate;
my $currLibrary;
my $prevLibrary;

# EACH TABLE IS SCANNED CLONE BY CLONE (after one clone has been read it must be processed)
# arrays for sizes (true and background), reused per each clone
# vector bands and wrong area should have already been removed
my @blueSizes = ();
my @greenSizes = ();
my @yellowSizes = ();
my @redSizes = ();
my %sizes = (
     B => \@blueSizes,
     G => \@greenSizes,
     Y => \@yellowSizes,
     R => \@redSizes
);

# arrays for heights (linked to Sizes arrays above), reused per each clone
my @blueHeights = ();
my @greenHeights = ();
my @yellowHeights = ();
my @redHeights = ();
my %heights = (
     B => \@blueHeights,
     G => \@greenHeights,
     Y => \@yellowHeights,
     R => \@redHeights
);

# working array to order heights of one color in one well, reused per each clone
my @sortedHeights = ();

# thresholds (per color)
my %threshold = (
     B => 0,
     G => 0,
     Y => 0,
     R => 0
);

# arrays for FPC converted sizes (only true bands), reused per each clone
my @FPCsizes = ();
my $FPCindex = 0;
my %FPCcolorIndex = (
     B => 0,
     G => 0,
     Y => 0,
     R => 0
);

# parameters for iterative method
# we order peaks in height and make high and low average of some peaks
my $highAVG; # initially avg of values between $firstValue and lastValue
# then avg values between $firstValue and $highLimit height
my $lowAVG; # initially avg of values from $lowIndex (see below) on
# then from $lowLimit (see further below) height down
my $avgRatio; # ratio between high and low avg values


# Tk, interface, initialization
my $mw = new MainWindow; # Main Window
$mw -> configure(-title=>"FPB - FingerPrint Background removal");

my $frm_name = $mw -> Frame();

my $save = $mw -> Button(-text=>"Save", -command =>\&save);
my $process = $mw -> Button(-text=>"Process", -command=>\&process);
my $vector = $mw -> Button(-text=>"Show vector", -command=>\&printVector);
my $quit = $mw -> Button(-text=>"Quit",- command=>sub{exit}); # quit button

my $firstValueLAB = $frm_name -> Label(-text=>"First Value:") -> grid(-row=>1,-column=>1);
my $firstValueENT = $frm_name -> Entry() -> grid(-row=>1,-column=>2);
my $lastValueLAB = $frm_name -> Label(-text=>"Last Value:") -> grid(-row=>1,-column=>3);
my $lastValueENT = $frm_name -> Entry() -> grid(-row=>1,-column=>4);
my $firstValue;
my $lastValue;

my $lowIndexLAB = $frm_name -> Label(-text=>"Low index:") -> grid(-row=>2,-column=>1);
my $lowIndexENT = $frm_name -> Entry() -> grid(-row=>2,-column=>2);
my $minBandsLAB = $frm_name -> Label(-text=>"Min bands:") -> grid(-row=>2,-column=>3);
my $minBandsENT = $frm_name -> Entry() -> grid(-row=>2,-column=>4);
my $lowIndex;
my $minBands;

my $minSizesPerColorLAB = $frm_name -> Label(-text=>"Min sizes (per color):") -> grid(-row=>3,-column=>1);#,-columnspan=>2,-sticky=>'e');
my $minSizesPerColorENT = $frm_name -> Entry() -> grid(-row=>3,-column=>2);
my $maxSizesLAB = $frm_name -> Label(-text=>"Max sizes (total):") -> grid(-row=>3,-column=>3);
my $maxSizesENT = $frm_name -> Entry() -> grid(-row=>3,-column=>4);

my $blueBackgroundLAB = $frm_name -> Label(-text=>"Blue background:") -> grid(-row=>4,-column=>1);
my $blueBackgroundENT = $frm_name -> Entry() -> grid(-row=>4,-column=>2);
my $greenBackgroundLAB = $frm_name -> Label(-text=>"Green background:") -> grid(-row=>4,-column=>3);
my $greenBackgroundENT = $frm_name -> Entry() -> grid(-row=>4,-column=>4);

my $yellowBackgroundLAB = $frm_name -> Label(-text=>"Yellow background:") -> grid(-row=>5,-column=>1);
my $yellowBackgroundENT = $frm_name -> Entry() -> grid(-row=>5,-column=>2);
my $redBackgroundLAB = $frm_name -> Label(-text=>"Red background:") -> grid(-row=>5,-column=>3);
my $redBackgroundENT = $frm_name -> Entry() -> grid(-row=>5,-column=>4);

my $blueOffsetLAB = $frm_name -> Label(-text=>"Blue offset:") -> grid(-row=>6,-column=>1);
my $blueOffsetENT = $frm_name -> Entry() -> grid(-row=>6,-column=>2);
my $greenOffsetLAB = $frm_name -> Label(-text=>"Green offset:") -> grid(-row=>6,-column=>3);
my $greenOffsetENT = $frm_name -> Entry() -> grid(-row=>6,-column=>4);

my $yellowOffsetLAB = $frm_name -> Label(-text=>"Yellow offset:") -> grid(-row=>7,-column=>1);
my $yellowOffsetENT = $frm_name -> Entry() -> grid(-row=>7,-column=>2);
my $redOffsetLAB = $frm_name -> Label(-text=>"Red offset:") -> grid(-row=>7,-column=>3);
my $redOffsetENT = $frm_name -> Entry() -> grid(-row=>7,-column=>4);

my $toleranceLAB = $frm_name -> Label(-text=>"Tolerance:") -> grid(-row=>8,-column=>1);
my $toleranceENT = $frm_name -> Entry() -> grid(-row=>8,-column=>2);
my $multiplyFactorLAB = $frm_name -> Label(-text=>"Multiply factor:") -> grid(-row=>8,-column=>3);
my $multiplyFactorENT = $frm_name -> Entry() -> grid(-row=>8,-column=>4);

my $peakWidthLAB = $frm_name -> Label(-text=>"Peak width:") -> grid(-row=>9,-column=>1);
my $peakWidthENT = $frm_name -> Entry() -> grid(-row=>9,-column=>2);
my $fixedThresholdLAB = $frm_name -> Label(-text=>"Fixed threshold:") -> grid(-row=>9,-column=>3);
my $fixedThresholdENT = $frm_name -> Entry() -> grid(-row=>9,-column=>4);

my $sizeFromLAB = $frm_name -> Label(-text=>"Size from:") -> grid(-row=>10,-column=>1);
my $sizeFromENT = $frm_name -> Entry() -> grid(-row=>10,-column=>2);
my $sizeToLAB = $frm_name -> Label(-text=>"Size to:") -> grid(-row=>10,-column=>3);
my $sizeToENT = $frm_name -> Entry() -> grid(-row=>10,-column=>4);

my $libFromLAB = $frm_name -> Label(-text=>"Library from:") -> grid(-row=>11,-column=>1);
my $libFromENT = $frm_name -> Entry() -> grid(-row=>11,-column=>2);
my $libToLAB = $frm_name -> Label(-text=>"Library to:") -> grid(-row=>11,-column=>3);
my $libToENT = $frm_name -> Entry() -> grid(-row=>11,-column=>4);

my $plateFromLAB = $frm_name -> Label(-text=>"Plate from:") -> grid(-row=>12,-column=>1);
my $plateFromENT = $frm_name -> Entry() -> grid(-row=>12,-column=>2);
my $plateToLAB = $frm_name -> Label(-text=>"Plate to:") -> grid(-row=>12,-column=>3);
my $plateToENT = $frm_name -> Entry() -> grid(-row=>12,-column=>4);

my $gridFromLAB = $frm_name -> Label(-text=>"Grid from:") -> grid(-row=>13,-column=>1);
my $gridFromENT = $frm_name -> Entry() -> grid(-row=>13,-column=>2);
my $tableSuffixLAB = $frm_name -> Label(-text=>"Table suffix:") -> grid(-row=>13,-column=>3);
my $tableSuffixENT = $frm_name -> Entry() -> grid(-row=>13,-column=>4);

#my $initialRatioLAB = $frm_name -> Label(-text=>"Initial ratio:") -> grid(-row=>14,-column=>1);
#my $initialRatioENT = $frm_name -> Entry() -> grid(-row=>14,-column=>2);

#my $thresholdMultiplierLAB = $frm_name -> Label(-text=>"Threshold multiplier:") -> grid(-row=>14,-column=>3);
#my $thresholdMultiplierENT = $frm_name -> Entry() -> grid(-row=>14,-column=>4);

my $logText = $mw -> Frame();
my $txt = $logText -> Text(-height=>25);
my $scrollBar = $logText -> Scrollbar(-orient=>'v',-command=>[yview => $txt]);
$txt -> configure(-yscrollcommand=>['set', $scrollBar]);

$frm_name -> grid(-row=>1,-column=>1,-columnspan=>4);
$save -> grid(-row=>15,-column=>1);
$process -> grid(-row=>15,-column=>2);
$vector -> grid(-row=>15,-column=>3);
$quit -> grid(-row=>15,-column=>4);

$logText -> grid(-row=>16,-column=>1,-columnspan=>4);
$txt -> grid(-row=>1,-column=>1);
$scrollBar -> grid(-row=>1,-column=>2,-sticky=>"ns");

# READ PARAMETERS.TXT FILE (IF IT EXISTS)
if (open(PARAM,"parameters.cfg")) {# or die("Can't open file with parameters from local directory\n");
  while($line=<PARAM>) {
    if ($line=~ /firstValue=(\d+)/) {$firstValueENT -> insert('end',$1);}
    if ($line=~ /lastValue=(\d+)/) {$lastValueENT -> insert('end',$1);}
    if ($line=~ /lowIndex=(\d+)/) {$lowIndexENT -> insert('end',$1);}

    # minimum number of bands per color (if fewer bands then fixedThreshold, see below)
    if ($line=~ /minBands=(\d+)/) {$minBandsENT -> insert('end',$1);}

    # minimum cutoff, we do not pick in any case peaks under this value
    if ($line=~ /blueBackground=(\d+)/) {$blueBackgroundENT -> insert('end',$1);}
    if ($line=~ /greenBackground=(\d+)/) {$greenBackgroundENT -> insert('end',$1);}
    if ($line=~ /yellowBackground=(\d+)/) {$yellowBackgroundENT -> insert('end',$1);}
    if ($line=~ /redBackground=(\d+)/) {$redBackgroundENT -> insert('end',$1);}
    if ($line=~ /tolerance=(\d+\.\d+)/) {$toleranceENT -> insert('end',$1);}
    # conversion to FPC sizes
    if ($line=~ /blueOffset=(\d+)/) {$blueOffsetENT -> insert('end',$1);}
    if ($line=~ /greenOffset=(\d+)/) {$greenOffsetENT -> insert('end',$1);}
    if ($line=~ /yellowOffset=(\d+)/) {$yellowOffsetENT -> insert('end',$1);}
    if ($line=~ /redOffset=(\d+)/) {$redOffsetENT -> insert('end',$1);}
    if ($line=~ /multiplyFactor=(\d+)/) {$multiplyFactorENT -> insert('end',$1);}
    # area/height <= peakWidth
    if ($line=~ /peakWidth=(\d+)/) {$peakWidthENT -> insert('end',$1);}
    if ($line=~ /sizeFrom=(\d+)/) {$sizeFromENT -> insert('end',$1);}
    if ($line=~ /sizeTo=(\d+)/) {$sizeToENT -> insert('end',$1);}
    if ($line=~ /libFrom=(\d+)/) {$libFromENT -> insert('end',$1);}
    if ($line=~ /libTo=(\d+)/) {$libToENT -> insert('end',$1);}
    if ($line=~ /plateFrom=(\d+)/) {$plateFromENT -> insert('end',$1);}
    if ($line=~ /plateTo=(\d+)/) {$plateToENT -> insert('end',$1);}
    if ($line=~ /gridFrom=(\d+)/) {$gridFromENT -> insert('end',$1);}
    if ($line=~ /fixedThreshold=(\d+)/) {$fixedThresholdENT -> insert('end',$1);}
    if ($line=~ /tableSuffix=(\w+)/) {$tableSuffixENT -> insert('end',$1);}
    if ($line=~ /minSizesPerColor=(\d+)/) {$minSizesPerColorENT -> insert('end',$1);}
    if ($line=~ /maxSizes=(\d+)/) {$maxSizesENT -> insert('end',$1);}
#    if ($line=~ /initialRatio=(\d+\.?\d*)/) {$initialRatioENT -> insert('end',$1);}
#    if ($line=~ /thresholdMultiplier=(\d+\.?\d*)/) {$thresholdMultiplierENT -> insert('end',$1);}

    #  if ($line=~ /gridTo=(\d+)/) {$gridToENT -> insert('end',$1);}
    #  if ($line=~ /libSizes=(\w+)/) {$libSizesENT -> insert('end',$1);} # used in size names for FPC
    #  if ($line=~ /libChecked=(true|false)/) {$libCheckedENT -> insert('end',$1);} # used in size names for FPC

  }
  close PARAM;
}

my @blueVector = ();
my @greenVector = ();
my @yellowVector = ();
my @redVector = ();
my %vector = (
     B => \@blueVector,
     G => \@greenVector,
     Y => \@yellowVector,
     R => \@redVector
);

MainLoop;

sub readVector {
  my $vectorOpen = open(VECTOR,"vector.cfg");
  if ($vectorOpen) {
    while($line=<VECTOR>) {
      if ($line =~ /Blue/) {
	while ($line2 = <VECTOR>) {
	  last if ($line2 =~ /-1/);
	  if ($line2 =~ /(\d+\.?\d?\d?)/) {
	    push(@{$vector{'B'}},$1);
	  }
	}
      }
      if ($line =~ /Green/) {
	while ($line2 = <VECTOR>) {
	  last if ($line2 =~ /-1/);
	  if ($line2 =~ /(\d+\.?\d?\d?)/) {
	    push(@{$vector{'G'}},$1);
	  }
	}
      }
      if ($line =~ /Yellow/) {
	while ($line2 = <VECTOR>) {
	  last if ($line2 =~ /-1/);
	  if ($line2 =~ /(\d+\.?\d?\d?)/) {
	    push(@{$vector{'Y'}},$1);
	  }
	}
      }
      if ($line =~ /Red/) {
	while ($line2 = <VECTOR>) {
	  last if ($line2 =~ /-1/);
	  if ($line2 =~ /(\d+\.?\d?\d?)/) {
	    push(@{$vector{'R'}},$1);
	  }
	}
      }
    }
    close VECTOR;
  }
  else { # vector not open
    $txt -> insert('end', "CANNOT OPEN VECTOR FILE vector.cfg!\n\n");
  }
  return $vectorOpen;
}

sub printVector {
  my $vectorOpen = open(VECTOR,"vector.cfg");
  if ($vectorOpen) {
    while($line=<VECTOR>) {
      $txt -> insert('end', $line);
    }
    close VECTOR;
    $txt -> insert('end', "\n\n");
  }
  else { # vector not open
    $txt -> insert('end', "CANNOT OPEN VECTOR FILE vector.cfg!\n\n");
  }
  return $vectorOpen;
}

sub save {
  if ((valuesOK()) &&
       ((open(PARAM,">parameters.cfg")) or (die("Cannot open parameters.cfg file\n")))) {
    print(PARAM "# Parameters in iterate.pl\n");
    print(PARAM "# Names correspond to variables\n");
    print(PARAM "firstValue=".$firstValueENT->get()."\n");
    print(PARAM "lastValue=".$lastValueENT->get()."\n");
    print(PARAM "lowIndex=".$lowIndexENT->get()."\n");
    print(PARAM "minBands=".$minBandsENT->get()."\n");
    print(PARAM "blueBackground=".$blueBackgroundENT->get()."\n");
    print(PARAM "greenBackground=".$greenBackgroundENT->get()."\n");
    print(PARAM "yellowBackground=".$yellowBackgroundENT->get()."\n");
    print(PARAM "redBackground=".$redBackgroundENT->get()."\n");
    print(PARAM "tolerance=".$toleranceENT->get()."\n");
    print(PARAM "blueOffset=".$blueOffsetENT->get()."\n");
    print(PARAM "greenOffset=".$greenOffsetENT->get()."\n");
    print(PARAM "yellowOffset=".$yellowOffsetENT->get()."\n");
    print(PARAM "redOffset=".$redOffsetENT->get()."\n");
    print(PARAM "multiplyFactor=".$multiplyFactorENT->get()."\n");
    print(PARAM "peakWidth=".$peakWidthENT->get()."\n");
    print(PARAM "sizeFrom=".$sizeFromENT->get()."\n");
    print(PARAM "sizeTo=".$sizeToENT->get()."\n");
    print(PARAM "libFrom=".$libFromENT->get()."\n");
    print(PARAM "libTo=".$libToENT->get()."\n");
    print(PARAM "plateFrom=".$plateFromENT->get()."\n");
    print(PARAM "plateTo=".$plateToENT->get()."\n");
    print(PARAM "gridFrom=".$gridFromENT->get()."\n");
    print(PARAM "fixedThreshold=".$fixedThresholdENT->get()."\n");
    print(PARAM "tableSuffix=".$tableSuffixENT->get()."\n");
    print(PARAM "minSizesPerColor=".$minSizesPerColorENT->get()."\n");
    print(PARAM "maxSizes=".$maxSizesENT->get()."\n");
#    print(PARAM "initialRatio=".$initialRatioENT->get()."\n");
#    print(PARAM "thresholdMultiplier=".$thresholdMultiplierENT->get()."\n");
    close PARAM;
    $txt -> insert('end', "DATA SAVED TO parameters.cfg\n\n");
  }
  else {
    $txt -> insert('end', "DATA NOT SAVED!\n\n");
  }
}

sub valuesOK {
  my $allSet = 1; # check that all values are initialized

  if (length($firstValueENT->get()) == 0) {
    $txt -> insert('end',"FIRST VALUE has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($firstValueENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"FIRST VALUE is not a positive integer number!\n");
    $allSet = 0;
  }
  elsif (length($lastValueENT->get()) == 0) {
    $txt -> insert('end',"LAST VALUE has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($lastValueENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"LAST VALUE is not a positive integer number!\n");
    $allSet = 0;
  }
  elsif ($lastValueENT->get() < $firstValueENT->get()) {
    $txt -> insert('end',"LAST VALUE lower than FIRST VALUE!\n");
    $allSet = 0;
  }
  elsif (length($minBandsENT->get()) == 0) {
    $txt -> insert('end',"MIN BANDS has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($minBandsENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"MIN BANDS is not a positive integer number!\n");
    $allSet = 0;
  }
  elsif ($minBandsENT->get() < $lastValueENT->get()) { # should be +5
    $txt -> insert('end',"MIN BANDS lower than LAST VALUE!\n");
    $allSet = 0;
  }
  elsif (length($lowIndexENT->get()) == 0) {
    $txt -> insert('end',"LOW INDEX has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($lowIndexENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"LOW INDEX is not a positive integer number!\n");
    $allSet = 0;
  }
  elsif ($lowIndexENT->get() < $minBandsENT->get()) {
    $txt -> insert('end',"LOW INDEX lower than MIN BANDS!\n");
    $allSet = 0;
  }

  if (length($blueBackgroundENT->get()) == 0) {
    $txt -> insert('end',"BLUE BACKGROUND has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($blueBackgroundENT->get() =~ /^(0|[1-9]\d*)$/)) {
    $txt -> insert('end',"BLUE BACKGROUND is not an integer number!\n");
    $allSet = 0;
  }

  if (length($greenBackgroundENT->get()) == 0) {
    $txt -> insert('end',"GREEN BACKGROUND has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($greenBackgroundENT->get() =~ /^(0|[1-9]\d*)$/)) {
    $txt -> insert('end',"GREEN BACKGROUND is not an integer number!\n");
    $allSet = 0;
  }

  if (length($yellowBackgroundENT->get()) == 0) {
    $txt -> insert('end',"YELLOW BACKGROUND has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($yellowBackgroundENT->get() =~ /^(0|[1-9]\d*)$/)) {
    $txt -> insert('end',"YELLOW BACKGROUND is not an integer number!\n");
    $allSet = 0;
  }

  if (length($redBackgroundENT->get()) == 0) {
    $txt -> insert('end',"RED BACKGROUND has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($redBackgroundENT->get() =~ /^(0|[1-9]\d*)$/)) {
    $txt -> insert('end',"RED BACKGROUND is not an integer number!\n");
    $allSet = 0;
  }

  if (length($blueOffsetENT->get()) == 0) {
    $txt -> insert('end',"BLUE OFFSET has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($blueOffsetENT->get() =~ /^(0|[1-9]\d*)$/)) {
    $txt -> insert('end',"BLUE OFFSET is not an integer number!\n");
    $allSet = 0;
  }

  if (length($greenOffsetENT->get()) == 0) {
    $txt -> insert('end',"GREEN OFFSET has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($greenOffsetENT->get() =~ /^(0|[1-9]\d*)$/)) {
    $txt -> insert('end',"GREEN OFFSET is not an integer number!\n");
    $allSet = 0;
  }

  if (length($yellowOffsetENT->get()) == 0) {
    $txt -> insert('end',"YELLOW OFFSET has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($yellowOffsetENT->get() =~ /^(0|[1-9]\d*)$/)) {
    $txt -> insert('end',"YELLOW OFFSET is not an integer number!\n");
    $allSet = 0;
  }

  if (length($redOffsetENT->get()) == 0) {
    $txt -> insert('end',"RED OFFSET has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($redOffsetENT->get() =~ /^(0|[1-9]\d*)$/)) {
    $txt -> insert('end',"RED OFFSET is not an integer number!\n");
    $allSet = 0;
  }

  if (length($toleranceENT->get()) == 0) {
    $txt -> insert('end',"TOLERANCE has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($toleranceENT->get() =~ /^(0|[1-9]\d*)(\.\d*[1-9])?$/)) { # possibly decimal number
    $txt -> insert('end',"TOLERANCE is not a positive decimal number!\n");
    $allSet = 0;
  }

  if (length($multiplyFactorENT->get()) == 0) {
    $txt -> insert('end',"MULTIPLY FACTOR has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($multiplyFactorENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"MULTIPLY FACTOR is not a positive integer number!\n");
    $allSet = 0;
  }

  if (length($peakWidthENT->get()) == 0) {
    $txt -> insert('end',"PEAK WIDTH has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($peakWidthENT->get() =~ /^(0|[1-9]\d*)(\.\d*[1-9])?$/)) {
    $txt -> insert('end',"PEAK WIDTH is not a positive rational number!\n");
    $allSet = 0;
  }

  if (length($fixedThresholdENT->get()) == 0) {
    $txt -> insert('end',"FIXED THRESHOLD has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($fixedThresholdENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"FIXED THRESHOLD is not a positive integer number!\n");
    $allSet = 0;
  }

  if (length($sizeFromENT->get()) == 0) {
    $txt -> insert('end',"SIZE FROM has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($sizeFromENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"SIZE FROM is not a positive integer number!\n");
    $allSet = 0;
  }
  elsif (length($sizeToENT->get()) == 0) {
    $txt -> insert('end',"SIZE TO has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($sizeToENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"SIZE TO is not a positive integer number!\n");
    $allSet = 0;
  }
  elsif ($sizeToENT->get() < $sizeFromENT->get()) {
    $txt -> insert('end',"SIZE TO lower than SIZE FROM!\n");
    $allSet = 0;
  }

  if (length($libFromENT->get()) == 0) {
    $txt -> insert('end',"LIBRARY FROM has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($libFromENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"LIBRARY FROM is not a positive integer number!\n");
    $allSet = 0;
  }
  elsif (length($libToENT->get()) == 0) {
    $txt -> insert('end',"LIBRARY TO has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($libToENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"LIBRARY TO is not a positive integer number!\n");
    $allSet = 0;
  }
  elsif ($libToENT->get() < $libFromENT->get()) {
    $txt -> insert('end',"LIBRARY TO lower than LIBRARY FROM!\n");
    $allSet = 0;
  }

  if (length($plateFromENT->get()) == 0) {
    $txt -> insert('end',"PLATE FROM has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($plateFromENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"PLATE FROM is not a positive integer number!\n");
    $allSet = 0;
  }
  elsif (length($plateToENT->get()) == 0) {
    $txt -> insert('end',"PLATE TO has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($plateToENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"PLATE TO is not a positive integer number!\n");
    $allSet = 0;
  }
  elsif ($plateToENT->get() < $plateFromENT->get()) {
    $txt -> insert('end',"PLATE TO lower than PLATE FROM!\n");
    $allSet = 0;
  }

  if (length($gridFromENT->get()) == 0) {
    $txt -> insert('end',"GRID FROM has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($gridFromENT->get() =~ /^[1-9]\d*$/)) {
    $txt -> insert('end',"GRID FROM is not a positive integer number!\n");
    $allSet = 0;
  }

  if (!($tableSuffixENT->get() =~ /^\w+$/)) {
    $txt -> insert('end',"TABLE SUFFIX is not a valid file suffix!\n");
    $allSet = 0;
  }

  if (length($minSizesPerColorENT->get()) == 0) {
    $txt -> insert('end',"MIN SIZES PER COLOR has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($minSizesPerColorENT->get() =~ /^\d+$/)) {
    $txt -> insert('end',"MIN SIZES PER COLOR is not a positive integer number!\n");
    $allSet = 0;
  }

  if (length($maxSizesENT->get()) == 0) {
    $txt -> insert('end',"INITIAL RATIO has not been set yet!\n");
    $allSet = 0;
  }
  elsif (!($maxSizesENT->get() =~ /^\d+$/)) {
    $txt -> insert('end',"Max Bands is not a positive integer number!\n");
    $allSet = 0;
  }

#  if (length($initialRatioENT->get()) == 0) {
#    $txt -> insert('end',"Initial Ratio has not been set yet!\n");
#    $allSet = 0;
# }
# elsif (!($initialRatioENT->get() =~ /^(\d+)(\.\d*[1-9])?$/)) { # possibly decimal number
#    $txt -> insert('end',"Initial Ratio is not a positive decimal number!\n");
#    $allSet = 0;
#  }

#  if (length($thresholdMultiplierENT->get()) == 0) {
#    $txt -> insert('end',"Threshold multiplier has not been set yet!\n");
#    $allSet = 0;
#  }
#  elsif (!($thresholdMultiplierENT->get() =~ /^(\d)(\.\d*[1-9])?$/)) { # possibly decimal number
#    $txt -> insert('end',"Threshold multiplier is not a positive decimal number!\n");
#    $allSet = 0;
#  }

  return $allSet;
}

sub process {
  $txt -> insert('end', "Start computing...\n");
  if ((!valuesOK()) || (!readVector())) {
    $txt -> insert('end', "CANNOT PROCESS!\n\n");
  }
  else {
    my $beforeLib = $libFromENT->get() - 1; # free characters before library starting position
    my $beforePlate = $plateFromENT->get() - ($libToENT->get() + 1); # free characters between library and plate
    my $beforeWell = $gridFromENT->get() - ($plateToENT->get() + 1); # free characters between plate and well
    my $inLib = ($libToENT->get() - $libFromENT->get()) + 1; # characters for library
    my $inPlate = ($plateToENT->get() - $plateFromENT->get()) + 1; # characters for plate

    my $suffix = $tableSuffixENT->get();

    $firstValue = $firstValueENT->get() - 1;
    $lastValue = $lastValueENT->get() - 1;
    $lowIndex = $lowIndexENT->get() - 1;
    $minBands = $minBandsENT->get() - 1;

    # list files in a directory
    # we pick only txt file which are tables ($item)
    @Dir = `ls`;
    foreach $item(@Dir){
      #chomp $item;
      if ($item =~ /(.*)\.$suffix$/) { # library and plate number required
	my $project = $1;

	# open table file (output of GeneMapper)
	#print "Table file name: ";
	my $table = $item;
	chomp $table;

	# read first line of data to initialize values
	open(TABLE,$table) or die("Can't open table file $table\n");
	$line = <TABLE>; # skip column description line
	$line = <TABLE>; # first line of data
	if ($line =~ /^"?([OBGYR]).*"?\s\w{$beforeLib}(\w{$inLib})\w{$beforePlate}(\w{$inPlate})\w{$beforeWell}([A-P][0-1]\d|[A-P]2[0-4])\w*\.(fsa|FSA)/) {
	  $prevColor = $1;
	  $prevLibrary = $2;
	  $prevPlate = $3;
	  $prevWell = $4;
	  close TABLE;
	}
	else {
	  print $line;
	  print "Wrong line parsing. Aborting.\n";
	  exit 1;
	}

	# reopen the table file to process it
	open(TABLE,$table) or die("Can't open table file $table\n");

	# debug files with threshold for each clone of a plate XXX names per plate!!!
	my $blueName = ">".$project."blueThreshold";
	open(B,$blueName) or die("Can't open debug file $blueName\n");
	my $greenName = ">".$project."greenThreshold";
	open(G,$greenName) or die("Can't open debug file $greenName\n");
	my $yellowName = ">".$project."yellowThreshold";
	open(Y,$yellowName) or die("Can't open debug file $yellowName\n");
	my $redName = ">".$project."redThreshold";
	open(R,$redName) or die("Can't open debug file $redName\n");

	my $FPCfile = $project . ".sizes";
	my $warning = $project . ".warning";
	open(FPCfile,">$FPCfile") or die("Can't open sizes file $FPCfile\n");
	open(WARNING,">$warning") or die("Can't open file $warning\n");

	# skip first line of input from the table file
	$line = <TABLE>;

	# Table is scanned to read all peaks and compute threshold (with extra array)
	# To add: check if clone is empty

	while ($line = <TABLE>) {
	  ################
	  # READING DATA #
	  ################
	  # reading is on CURRent data!
	  if ($line =~ /^"?([OBGYR])/) {
	    $currColor = $1;
	  }

	  #if (($line =~ /A01/) &&($currColor eq "G")) {
	  #  for($i = 0; $i <= $#{$sizes{$prevColor}}; $i++) {
	  #    print $sizes{$prevColor}->[$i]."\n";
	  #  }
	  #}

	  if ($currColor ne "O") {readData()};

	  ###################
	  # PROCESSING DATA #
	  ###################
	  # read dye, library, plate and well: if changed -> computation/iterations
	  # processing is on PREVious data!
	  if ($line =~ /\s\w{$beforeLib}(\w{$inLib})\w{$beforePlate}(\w{$inPlate})\w{$beforeWell}([A-P][0-1]\d|[A-P]2[0-4])\w*\.(fsa|FSA)/) {
	    $currLibrary = $1;
	    $currPlate = $2;
	    $currWell = $3;

	    # if something has changed you can process a new dye/well
	    if (($currColor ne $prevColor) || ($currLibrary ne $prevLibrary) || ($currPlate ne $prevPlate) || ($currWell ne $prevWell)) {
	      if ($prevColor ne "O") {
		processColor();
		addFPCsizes();
	      }
	    }

	    if (($currLibrary ne $prevLibrary) || ($currPlate ne $prevPlate) || ($currWell ne $prevWell)) {
	      printFPCsizes();
	      # empty data structures
	      @blueSizes = ();
	      @blueHeights = ();
	      @greenSizes = ();
	      @greenHeights = ();
	      @yellowSizes = ();
	      @yellowHeights = ();
	      @redSizes = ();
	      @redHeights = ();
	      @FPCsizes = ();
	      $FPCindex = 0;
	      $FPCcolorIndex{'B'} = 0; $FPCcolorIndex{'G'} = 0;
	      $FPCcolorIndex{'Y'} = 0; $FPCcolorIndex{'R'} = 0;
	    }
	  } # end if (changes)

	  # record info on data processed
	  $prevColor = $currColor;
	  $prevLibrary = $currLibrary;
	  $prevPlate = $currPlate;
	  $prevWell = $currWell;
	} # end_while (reading lines from table)

	# process last dye read and prind sizes for last well
	if ($prevColor ne "O") {
	  processColor();
	  addFPCsizes();
	}
	printFPCsizes();


	@blueSizes = ();
	@blueHeights = ();
	@greenSizes = ();
	@greenHeights = ();
	@yellowSizes = ();
	@yellowHeights = ();
	@redSizes = ();
	@redHeights = ();
	@FPCsizes = ();
	$FPCindex = 0;
	$FPCcolorIndex{'B'} = 0; $FPCcolorIndex{'G'} = 0;
	$FPCcolorIndex{'Y'} = 0; $FPCcolorIndex{'R'} = 0;

	close TABLE;
	close B;
	close G;
	close Y;
	close R;
	close FPCfile;

	$txt -> insert('end', "$table done!\n");
      } # end_if ($item =~ /.{$beforeLib}(.{$inLib}).{$beforePlate}(.{$inPlate}).*\.$suffix$/)
    } # end_forEach

    $txt -> insert('end', "\nAll files processed\n\n");



  }
}



sub readData() {
  if ($line =~ /^"?$currColor.*"?\s+.*\.(FSA|fsa)\s+(\d+\.\d+)\t(\d+)\t(\d+)/) {
    # assumption: tables have 4 columns of data, size, height, area, dataPoint
    my $size = $2;
    my $height = $3;
    my $area = $4;

 #   if (($line =~ /A01/) &&($currColor eq "G")) {
 #     for($i = 0; $i <= $#{$sizes{$prevColor}}; $i++) {
#	print "$currColor $prevColor: ".$sizes{$prevColor}->[$i]."\n";
#      }
#    }

    if (($size>=$sizeFromENT->get()) && ($size<=$sizeToENT->get()) && (($area/$height)<$peakWidthENT->get())) {
      my $i = 0; # vector check
      my $vectorBand = 0; # band not from vector
      my $vectBands = $#{$vector{$currColor}}; # number of vector bands
      while ($i <= $vectBands) {# da fare in modo furbo, incrementale!!! XXX
	if (($vector{$currColor}->[$i] - ($toleranceENT->get()/2) <= $size) && ($size <= $vector{$currColor}->[$i] + ($toleranceENT->get()/2))) {
	  $vectorBand=1; # size was within the tolerance
	}
	$i++;
      }
      if ($vectorBand == 0) { # if it is not a vector band push it in the sizes array
	push(@{$sizes{$currColor}},$size);
	push(@{$heights{$currColor}},$height);
      }
    }
  }
  else {
    #print "Wrong line: ".$line;
  }
}



sub processColor() {
  if ($#{$sizes{$prevColor}} < 0) {
    print(WARNING "$prevColor color totally absent in $prevLibrary".$prevPlate."_$prevWell\n");
  } else { # we have previous color already PREprocessed (and at least one band)
    @sortedHeights = sort sortNumbersDesc @{$heights{$prevColor}};

    my $minBands = $minBandsENT->get() - 1;
    if ($#sortedHeights <= $minBands) {
      $threshold{$prevColor} = $fixedThresholdENT->get();
    }
    else {
      $highAVG = 0;
      for (my $indexH = $firstValue; $indexH <= $lastValue; $indexH++) {
	$highAVG = $highAVG + $sortedHeights[$indexH];
      }
      $highAVG = $highAVG / ($lastValue - $firstValue + 1);

      $lowAVG = 0;
      my $startLowIndex;
      if ($#sortedHeights <= $lowIndex) {$startLowIndex = $#sortedHeights - 4;} # last 5 peaks
      else {$startLowIndex = $lowIndex;} # from $lowIndex down

      for (my $indexL = $startLowIndex; $indexL <= $#sortedHeights; $indexL++) {
	$lowAVG = $lowAVG + $sortedHeights[$indexL];
      }
      $lowAVG = $lowAVG / ($#sortedHeights - $startLowIndex + 1);
      $avgRatio = $highAVG / $lowAVG;

      my $highLimit = $sortedHeights[$lastValue];
      my $lowLimit = $sortedHeights[$startLowIndex];
      my $previousHighLimit = $sortedHeights[$lastValue];
      my $previousLowLimit = $sortedHeights[$startLowIndex];
      my $previousHighAVG = $highAVG + 1; # set in an odd way just to enter while loop below
      my $previousLowAVG = $lowAVG - 1;
      my $initialGap = $highAVG - $lowAVG;

      # if the two average values are close to each other at the beginning then do not iterate
      # just keep a few of them (proportionally to the avgRatio)
      my $initialRatio = 5;#$initialRatioENT->get();
      if ($avgRatio <= $initialRatio) {
	# keep a number of peaks proportional to the avgRatio
	my $difference = $avgRatio - 1; # avgRatio=1 is the lower bound
	my $percentage = ($initialRatio - 1) / $lastValue; # linear proportion
	my $keep = int($difference/$percentage);
	$highLimit = $sortedHeights[$keep];
      }
      else {
	while (($previousHighAVG > $highAVG) && (int($highLimit)>int($lowLimit))) { # (gap not reached) && (limits have not been met yet)
	  $previousHighAVG = $highAVG;
	  $previousLowAVG = $lowAVG;

	  my $diffLimit = $highLimit - $lowLimit;
	  my $diffAVG = $highAVG - $lowAVG;
	  my $displacement = (0.1 * $diffLimit + 0.1 * $diffAVG) * ($diffLimit/$diffAVG) + 0.1 * $initialGap;
	  #		my $avgLimit = $highLimit / $lowLimit;
	  #		my $displacement = $param1 * $initialGap;
	  #my $displacement = ($param1 * $avgLimit + $param2 * $avgRatio) * ($avgLimit/$avgRatio) + $param3 * $initialGap;
	  $highLimit = $highLimit - $displacement;
	  $lowLimit = $lowLimit + $displacement / 5;

	  if ($highLimit < $lowLimit) {$highLimit = $lowLimit;}#{$highLimit = $previousHighLimit; $lowLimit = $previousLowLimit;}
	  if ($previousHighLimit < $highLimit) {$highLimit = $previousHighLimit;}
	  if ($previousLowLimit > $lowLimit) {$lowLimit = $previousLowLimit;}
	  $previousHighLimit = $highLimit;
	  $previousLowLimit = $lowLimit;

	  my $index = $firstValue;
	  $highAVG = 0;
	  while ($sortedHeights[$index] >= $highLimit) {
	    $highAVG = $highAVG + $sortedHeights[$index];
	    $index++;
	  }
	  $highAVG = $highAVG / ($index - $firstValue);

	  $index = $#sortedHeights;
	  $lowAVG = 0;
	  while ($sortedHeights[$index] < $lowLimit) {
	    $lowAVG = $lowAVG + $sortedHeights[$index];
	    $index--;
	  }
	  $lowAVG = $lowAVG / ($#sortedHeights - $index);
	  $avgRatio = $highAVG / $lowAVG;
	}
      }

      $threshold{$prevColor} = int($highLimit) * 1;#$thresholdMultiplierENT->get();

      if ($prevColor eq "B") {
	if ($threshold{$prevColor} < $blueBackgroundENT->get()) {$threshold{$prevColor} = $blueBackgroundENT->get();}
      }
      elsif ($prevColor eq "G") {
	if ($threshold{$prevColor} < $greenBackgroundENT->get()) {$threshold{$prevColor} = $greenBackgroundENT->get();}
      }
      elsif ($prevColor eq "Y") {
	if ($threshold{$prevColor} < $yellowBackgroundENT->get()) {$threshold{$prevColor} = $yellowBackgroundENT->get();}
      }
      elsif ($prevColor eq "R") {
	if ($threshold{$prevColor} < $redBackgroundENT->get()) {$threshold{$prevColor} = $redBackgroundENT->get();}
      }
    }
    @sortedHeights = ();

    print($prevColor $prevLibrary.$prevPlate."_$prevWell:\t$threshold{$prevColor}\n");
  } # end PREVious dye analysis
}

sub addFPCsizes() {
  # print true bands to FPCfile (.sizes)
  my $FPCvalue;
  my $i;
  for($i = 0; $i <= $#{$sizes{$prevColor}}; $i++) {
    if ($heights{$prevColor}->[$i] >= $threshold{$prevColor}) {
      if ($prevColor eq "B") {
	$FPCvalue = $sizes{$prevColor}->[$i] * $multiplyFactorENT->get() + $blueOffsetENT->get();
      }
      elsif ($prevColor eq "G") {
	$FPCvalue = $sizes{$prevColor}->[$i] * $multiplyFactorENT->get() + $greenOffsetENT->get();
      }
      elsif ($prevColor eq "Y") {
	$FPCvalue = $sizes{$prevColor}->[$i] * $multiplyFactorENT->get() + $yellowOffsetENT->get();
      }
      elsif ($prevColor eq "R") {
	$FPCvalue = $sizes{$prevColor}->[$i] * $multiplyFactorENT->get() + $redOffsetENT->get();
      }
      $FPCsizes[$FPCindex] = $FPCvalue;
      $FPCindex++;
      $FPCcolorIndex{$prevColor}++;
    }
  }
}

sub printFPCsizes() {
  if ($FPCindex > $maxSizesENT->get()) {
    print (WARNING "Too many bands ($FPCindex) in ".$prevLibrary.$prevPlate."_$prevWell clone discarded\n");
    return;
  }
  else {
    my $discard = 0;
    foreach $key(keys(%FPCcolorIndex)) {
      if ($FPCcolorIndex{$key} < $minSizesPerColorENT->get()) {
	print (WARNING "Low number of sizes in color $key of ".$prevLibrary.$prevPlate."_$prevWell, clone discarded\n");
	$discard = 1;
      }
    }
    if ($discard == 1) {
      return;# exits subroutine
    }
    # min number of sizes reached for all colors
    my $exclude = $FPCindex - 4; # first band for each dye is removed (it can be a dye blob)
    print(FPCfile $prevLibrary.$prevPlate."_".$prevWell."\t$exclude\tGel\n");
    for($i=1; $i < $FPCcolorIndex{'B'}; $i++) { # exclude first band which can be a dye Blob
      print(FPCfile "$FPCsizes[$i]\n");
    }
    for($i=$FPCcolorIndex{'B'}+1; $i < $FPCcolorIndex{'B'}+$FPCcolorIndex{'G'}; $i++) { # exclude first band which can be a dye Blob
      print(FPCfile "$FPCsizes[$i]\n");
    }
    for($i=$FPCcolorIndex{'B'}+$FPCcolorIndex{'G'}+1; $i < $FPCcolorIndex{'B'}+$FPCcolorIndex{'G'}+$FPCcolorIndex{'Y'}; $i++) { # exclude first band which can be a dye Blob
      print(FPCfile "$FPCsizes[$i]\n");
    }
    for($i=$FPCcolorIndex{'B'}+$FPCcolorIndex{'G'}+$FPCcolorIndex{'Y'}+1; $i < $FPCcolorIndex{'B'}+$FPCcolorIndex{'G'}+$FPCcolorIndex{'Y'}+$FPCcolorIndex{'R'}; $i++) { # exclude first band which can be a dye Blob
      print(FPCfile "$FPCsizes[$i]\n");
    }
    print(FPCfile "-1\n");
  }
}



sub sortNumbersDesc { $b <=> $a }
